/*
Cyber Knight 2: LZ Compression
*/

#include <stdio.h>
#include <memory.h>

//#define DEBUG_LZ

unsigned char cache[0x1000];
unsigned short cache_ptr;

unsigned char buffer[0x10000];
unsigned short buf_ptr;

int cache_mask;
int run_mask;
int win_mask;
int win_bits;

///////////////////////////////////////////////////////

void LZ_Decode( FILE *fp, FILE *out, int offset, int length )
{
	int pattern;
	int bits;

	// init
	buf_ptr = 0;
	bits = 0;
	memset( cache,0,sizeof( cache ) );

	// start of data
	//fseek( fp, offset, SEEK_SET );

	// check decompression parameters
	switch( fgetc( fp ) ) {
		case 0:
			run_mask = 0x0f;
			win_mask = 0xf0;
			win_bits = 4;
			cache_ptr = 0xfee;
			cache_mask = 0xfff;
			break;

		case 1:
			run_mask = 0x1f;
			win_mask = 0xe0;
			win_bits = 3;
			cache_ptr = 0x7de;
			cache_mask = 0x7ff;
			break;

		case 2:
			run_mask = 0x3f;
			win_mask = 0xc0;
			win_bits = 2;
			cache_ptr = 0x3be;
			cache_mask = 0x3ff;
			break;

		case 3:
			run_mask = 0x7f;
			win_mask = 0x80;
			win_bits = 1;
			cache_ptr = 0x17e;
			cache_mask = 0x1ff;
			break;

		// memory copy
		case 4:
			fread( buffer, 1, length-1, fp );
			fwrite( buffer, 1, length-1, out );
			return;
			break;
	}

	length++;
	while( length ) {
		// grab compression usage
		if( bits == 0 ) {
			pattern = fgetc( fp );
			bits = 8;

			length--; if( !length ) break;
		}

		// Raw byte copy
		if( pattern & 1 ) {
			buffer[ buf_ptr ] = fgetc( fp );
			cache[ cache_ptr & cache_mask ] = buffer[ buf_ptr ];
		
			length--; if( !length ) break;
			
			buf_ptr++;
			cache_ptr++; cache_ptr &= cache_mask;
		}
		// LZ copy
		else {
			short win, run;

			// grab parameters
			win = fgetc( fp );
			length--; if( !length ) break;

			run = fgetc( fp );
			length--; if( !length ) break;

			// unpack upper window bits
			win |= ( run & win_mask ) << win_bits;

			// unpack run bits + overhead
			run = ( run & run_mask ) + 3;

			// copy for size of LZ window
			while( run-- ) {
				buffer[ buf_ptr ] = cache[ win & cache_mask ];
				cache[ cache_ptr & cache_mask ] = buffer[ buf_ptr ];

				buf_ptr++;
				win++; win &= cache_mask;
				cache_ptr++; cache_ptr &= cache_mask;
			}
		} // end decode

		// consume 1 bit
		pattern >>= 1;
		bits--;
	} // stop full decoder

#ifdef DEBUG_LZ
	FILE *debug = fopen( "debug.bin", "wb" );
	fwrite( buffer, 1, buf_ptr, debug );
	fclose( debug );
#endif

	fwrite( buffer, 1, buf_ptr, out );
}